import sys
from logging import Formatter, StreamHandler, getLogger
from logging import Logger as _Logger

DEBUG = 10
INFO = 20
WARNING = 30
ERROR = 40
LEVELS = {"DEBUG": DEBUG, "INFO": INFO, "WARNING": WARNING, "ERROR": ERROR}


class Logger:
    """Basic logger."""

    def __init__(self, name: str):
        self.name = name

        # return a logger with the specified name, creating it if necessary
        self._logger = getLogger(name)
        self._formatter = Formatter("[%(asctime)s] [%(levelname)s] %(message)s")

        # stream handler to stdout
        self._stream_handler = StreamHandler(sys.stdout)
        self._stream_handler.setFormatter(self._formatter)
        self._logger.addHandler(self._stream_handler)

        # set level of both the logger and the handler to INFO by default
        self.set_level("INFO")

    def set_level(self, level: str | int):
        """Set log level.

        Parameters
        ----------
        level : str | int
            Level to set.
        """
        # translate string to integer
        if isinstance(level, str):
            level = LEVELS[level.upper()]

        # set the logger's level
        self._logger.setLevel(level)

        # and all handlers
        for handler in self._logger.handlers:
            handler.setLevel(level)

    def debug(self, message: str):
        """Debug log message.

        Parameters
        ----------
        message : str
            Message to log.
        """
        if self._logger.isEnabledFor(DEBUG):
            self._logger._log(DEBUG, message, ())

    def info(self, message: str):
        """Info log message.

        Parameters
        ----------
        message : str
            Message to log
        """
        if self._logger.isEnabledFor(INFO):
            self._logger._log(INFO, message, ())

    def warning(self, message: str):
        """Warning log message.

        Parameters
        ----------
        message : str
            Message to log.
        """
        if self._logger.isEnabledFor(WARNING):
            self._logger._log(WARNING, message, ())

    def error(self, message: str | Exception):
        """Error log message.

        Parameters
        ----------
        message : str
            Message to log.
        """
        if self._logger.isEnabledFor(ERROR):
            if isinstance(message, Exception):
                # log stacktrace if message is an exception
                self._logger._log(ERROR, message, (), exc_info=True)
            else:
                self._logger._log(ERROR, message, ())

    @property
    def in_debug_mode(self) -> bool:
        """Checks if the logger's level is DEBUG.

        Returns
        -------
        bool
            True, if logger is in DEBUG mode, False otherwise.
        """
        return self._logger.level == DEBUG

    @property
    def level(self) -> int:
        """Returns the current log level.

        Returns
        -------
        int
            Log level.
        """
        return self._logger.level

    def __repr__(self):
        """Representation of the logger."""
        return f"<Logger: {self.name} ({self.level})>"


def get_logger(name: str) -> _Logger:
    """Get the specified logger object.

    Parameters
    ----------
    name : str
        Name of the module to get the logger for.

    Returns
    -------
    _Logger
        Logger of the specified module.
    """
    return getLogger(name)


def logger_exists(name: str) -> bool:
    """Check if a logger exists for the specified module.

    Parameters
    ----------
    name : str
        Name of the module to check the logger for.

    Returns
    -------
    bool
        True if logger exists, False otherwise.
    """
    return getLogger(name).hasHandlers()


def set_level(name: str, level: int | str):
    """Set log level for the specified logger.

    Parameters
    ----------
    name : str
        Name of the module.
    level : int | str
        Level to set.
    """
    logger = getLogger(name)

    # translate string to integer
    if isinstance(level, str):
        level = LEVELS[level.upper()]

    # set the logger's level
    logger.setLevel(level)

    # and all handlers
    for handler in logger.handlers:
        handler.setLevel(level)



from collections.abc import Generator

from mure.cache import Cache
from mure.iterator import ResponseIterator
from mure.models import Request, Resource, Response


def delete(
    resources: list[Resource],
    *,
    batch_size: int = 5,
    cache: Cache | None = None,
) -> Generator[Response, None, None]:
    """Perform a DELETE request for each resource.

    Parameters
    ----------
    resources : list[Resource]
        Resources to request.
    batch_size : int
        Number of items to request per batch concurrently, by default 5.
    cache : Cache | None, optional
        Cache to use for storing responses, by default None.

    Returns
    -------
    Generator[Response, None, None]
        The server's responses for each resource.
    """
    return (
        response
        for response in ResponseIterator(
            [Request("DELETE", **resource) for resource in resources],
            batch_size=batch_size,
            cache=cache,
        )
    )


def get(
    resources: list[Resource],
    *,
    batch_size: int = 5,
    cache: Cache | None = None,
) -> Generator[Response, None, None]:
    """Perform a GET request for each resource.

    Parameters
    ----------
    resources : list[Resource]
        Resources to request.
    batch_size : int
        Number of items to request per batch concurrently, by default 5.
    cache : Cache | None, optional
        Cache to use for storing responses, by default None.

    Returns
    -------
    Generator[Response, None, None]
        The server's responses for each resource.
    """
    return (
        response
        for response in ResponseIterator(
            [Request("GET", **resource) for resource in resources],
            batch_size=batch_size,
            cache=cache,
        )
    )


def head(
    resources: list[Resource],
    *,
    batch_size: int = 5,
    cache: Cache | None = None,
) -> Generator[Response, None, None]:
    """Perform a HEAD request for each resource.

    Parameters
    ----------
    resources : list[Resource]
        Resources to request.
    batch_size : int
        Number of items to request per batch concurrently, by default 5.
    cache : Cache | None, optional
        Cache to use for storing responses, by default None.

    Returns
    -------
    Generator[Response, None, None]
        The server's responses for each resource.
    """
    return (
        response
        for response in ResponseIterator(
            [Request("HEAD", **resource) for resource in resources],
            batch_size=batch_size,
            cache=cache,
        )
    )


def patch(
    resources: list[Resource],
    *,
    batch_size: int = 5,
    cache: Cache | None = None,
) -> Generator[Response, None, None]:
    """Perform a PATCH request for each resource.

    Parameters
    ----------
    resources : list[Resource]
        Resources to request.
    batch_size : int
        Number of items to request per batch concurrently, by default 5.
    cache : Cache | None, optional
        Cache to use for storing responses, by default None.

    Returns
    -------
    Generator[Response, None, None]
        The server's responses for each resource.
    """
    return (
        response
        for response in ResponseIterator(
            [Request("PATCH", **resource) for resource in resources],
            batch_size=batch_size,
            cache=cache,
        )
    )


def post(
    resources: list[Resource],
    *,
    batch_size: int = 5,
    cache: Cache | None = None,
) -> Generator[Response, None, None]:
    """Perform a POST request for each resource.

    Parameters
    ----------
    resources : list[Resource]
        Resources to request.
    batch_size : int
        Number of items to request per batch concurrently, by default 5.
    cache : Cache | None, optional
        Cache to use for storing responses, by default None.

    Returns
    -------
    Generator[Response, None, None]
        The server's responses for each resource.
    """
    return (
        response
        for response in ResponseIterator(
            [Request("POST", **resource) for resource in resources],
            batch_size=batch_size,
            cache=cache,
        )
    )


def put(
    resources: list[Resource],
    *,
    batch_size: int = 5,
    cache: Cache | None = None,
) -> Generator[Response, None, None]:
    """Perform a PUT request for each resource.

    Parameters
    ----------
    resources : list[Resource]
        Resources to request.
    batch_size : int
        Number of items to request per batch concurrently, by default 5.
    cache : Cache | None, optional
        Cache to use for storing responses, by default None.

    Returns
    -------
    Generator[Response, None, None]
        The server's responses for each resource.
    """
    return (
        response
        for response in ResponseIterator(
            [Request("PUT", **resource) for resource in resources],
            batch_size=batch_size,
            cache=cache,
        )
    )



from enum import StrEnum

from hishel import (
    HEURISTICALLY_CACHEABLE_STATUS_CODES,
    AsyncBaseStorage,
    AsyncFileStorage,
    AsyncInMemoryStorage,
    AsyncSQLiteStorage,
    Controller,
)


class Cache(StrEnum):
    """Available cache backends."""

    FILE = "FILE"
    IN_MEMORY = "IN_MEMORY"
    SQLITE = "SQLITE"


class CacheController(Controller):
    """Cache controller for HTTP requests."""

    def __init__(self):
        super().__init__(
            cacheable_methods=["GET", "POST"],
            force_cache=True,
            allow_heuristics=True,
            cacheable_status_codes=[*HEURISTICALLY_CACHEABLE_STATUS_CODES, 0, 500],
        )


def get_storage(cache: Cache) -> AsyncBaseStorage:
    """Get storage backend based on the cache type.

    Parameters
    ----------
    cache : Cache
        Cache type.

    Returns
    -------
    AsyncBaseStorage
        Storage backend.
    """
    match cache:
        case Cache.FILE:
            return AsyncFileStorage()
        case Cache.IN_MEMORY:
            return AsyncInMemoryStorage()
        case Cache.SQLITE:
            return AsyncSQLiteStorage()



from json import JSONDecodeError

import pytest

import mure
from mure.models import Resource, Response


@pytest.mark.xfail(reason="httpbin.org is sometimes down")
def test_get():
    resources: list[Resource] = [
        {"url": "https://httpbin.org/get"},
        {"url": "https://httpbin.org/get", "params": {"foo": "bar"}},
        {"url": "invalid"},
    ]

    responses: list[Response] = list(mure.get(resources, batch_size=2))

    assert len(responses) == 3
    assert responses[0].ok
    assert responses[1].ok
    assert not responses[2].ok


@pytest.mark.xfail(reason="httpbin.org is sometimes down")
def test_post():
    resources: list[Resource] = [
        {"url": "https://httpbin.org/post"},
        {"url": "https://httpbin.org/post", "json": {"foo": "bar"}},
        {"url": "invalid"},
    ]

    responses: list[Response] = list(mure.post(resources, batch_size=2))

    assert len(responses) == 3
    assert responses[0].ok
    assert responses[1].ok
    assert not responses[2].ok


@pytest.mark.xfail(reason="httpbin.org is sometimes down")
def test_json():
    response = next(mure.get([{"url": "https://httpbin.org/get"}]))

    assert len(response.json()) == 4


def test_invalid_json():
    with pytest.raises(JSONDecodeError):
        next(mure.get([{"url": "https://wikipedia.org"}])).json()



import asyncio
import contextlib
import os
from asyncio import AbstractEventLoop, Event, PriorityQueue, Task
from collections.abc import AsyncGenerator, Iterator
from typing import Self

import chardet
import httpcore
from hishel import AsyncCacheClient
from hishel._utils import generate_key
from httpx import AsyncClient

from mure.cache import Cache, CacheController, get_storage
from mure.logging import Logger
from mure.models import Request, Response

LOGGER = Logger(__name__)


class ResponseIterator(Iterator[Response]):
    """Response iterator that fetches responses concurrently."""

    def __init__(
        self,
        requests: list[Request],
        *,
        batch_size: int = 5,
        cache: Cache | None = None,
    ):
        """Initialize a response iterator.

        Parameters
        ----------
        requests : list[Request]
            Resources to request.
        batch_size : int, optional
            Number of resources to request concurrently, by default 5.
        cache : Cache | None, optional
            Cache to use for storing responses, by default None.
        """
        self.requests = requests
        self.num_requests = len(requests)
        self.pending = len(requests)
        self.batch_size = batch_size

        self._storage = get_storage(cache) if cache else None
        self._log_errors = bool(os.environ.get("MURE_LOG_ERRORS"))
        self._queue = PriorityQueue()
        self._events = [Event() for _ in requests]
        self._tasks: dict[int, Task] = {}
        self._responses = self._fetch_responses()

    def __repr__(self) -> str:
        """Response iterator representation.

        Returns
        -------
        str
            Representation with number of pending requests.
        """
        return f"<ResponseIterator: {self.pending}/{self.num_requests} pending>"

    def __len__(self) -> int | float:
        """Return the number of pending responses.

        Returns
        -------
        int
            Absolute number of pending responses.
        """
        return self.pending

    def __iter__(self) -> Self:
        """Yield one response at a time.

        Yields
        ------
        Iterator[Response]
            Response iterator.
        """
        return self

    def __next__(self) -> Response:
        """Return the next response.

        Returns
        -------
        Response
            Next response.
        """
        return next(self._responses)

    def _fetch_responses(self) -> Iterator[Response]:
        """Fetch responses concurrently.

        Yields
        ------
        Response
            One response at a time.
        """
        # get new event loop that is used for all operations
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        try:
            agenerator = self._agenerator_wrapper(loop)

            while True:
                try:
                    yield loop.run_until_complete(anext(agenerator))
                except StopAsyncIteration:
                    break
        except GeneratorExit:
            self._cancel_tasks()

            # run the event loop briefly to process task cancellations
            loop.run_until_complete(asyncio.sleep(0.5))
            raise
        finally:
            if not loop.is_closed():
                self._cancel_tasks()

                # run the event loop briefly to process task cancellations
                loop.run_until_complete(asyncio.sleep(0.5))
                loop.close()

            asyncio.set_event_loop(None)

    def _cancel_tasks(self):
        """Cancel all pending tasks."""
        for task in self._tasks.values():
            if not task.done() and not task.cancelled():
                task.cancel()

    def _schedule_tasks(self, session: AsyncClient, loop: AbstractEventLoop):
        """Schedule tasks for fetching responses concurrently.

        Parameters
        ----------
        session : AsyncClient
            HTTP session to use.
        loop : AbstractEventLoop
            Event loop to use.
        """
        for priority, request in enumerate(self.requests):
            coroutine = self._aprocess_request(
                session,
                priority,
                request,
                self._events[priority],
            )

            self._tasks[priority] = loop.create_task(coroutine)
            yield

    async def _aprocess_request(
        self,
        session: AsyncClient,
        priority: int,
        request: Request,
        event: Event,
    ):
        """Process a request by fetching and putting it in the queue.

        Parameters
        ----------
        session : AsyncClient
            HTTP session to use.
        priority : int
            Priority of the request.
        request. : Request
            Resource to request.
        event : Event
            Event to set when the response is ready.
        """
        LOGGER.debug(f"Started {priority}")

        response = await self._asend_request(session, request)

        # put response in the queue
        await self._queue.put((priority, response))

        # set event to notify that the response is ready
        event.set()

        LOGGER.debug(f"Finished {priority}")

    async def _agenerator_wrapper(self, loop: AbstractEventLoop) -> AsyncGenerator[Response, None]:
        """Wrap the response generator.

        Parameters
        ----------
        loop : AbstractEventLoop
            Event loop to use.

        Yields
        ------
        Response
            The server's response.
        """
        async for response in self._afetch_responses(loop):
            yield response

    async def _afetch_responses(self, loop: AbstractEventLoop) -> AsyncGenerator[Response, None]:
        """Fetch responses concurrently.

        Parameters
        ----------
        loop : AbstractEventLoop
            Event loop to use.

        Yields
        ------
        Response
            The server's response.
        """
        try:
            async with (
                AsyncClient(follow_redirects=True, http2=True)
                if self._storage is None
                else AsyncCacheClient(
                    follow_redirects=True,
                    http2=True,
                    storage=self._storage,
                    controller=CacheController(),
                )
            ) as session:
                # schedule tasks for fetching responses concurrently
                tasks = self._schedule_tasks(session, loop)
                while len(self._tasks) < self.batch_size:
                    try:
                        next(tasks)
                    except StopIteration:
                        break

                for event in self._events:
                    # wait for the specific event to be set to preserve order of the requests
                    await event.wait()

                    # get response from the queue
                    priority, response = await self._queue.get()

                    # get rid of the task that has been completed
                    self._tasks.pop(priority)

                    LOGGER.debug(f"Yielding {priority}")
                    yield response
                    self.pending -= 1

                    self._queue.task_done()

                    with contextlib.suppress(StopIteration):
                        # schedule next task (if any left)
                        next(tasks)
        except GeneratorExit:
            return
        finally:
            await asyncio.sleep(0.5)
            await self._queue.join()

    async def _asend_request(self, session: AsyncClient, request: Request) -> Response:
        """Perform a HTTP request.

        Parameters
        ----------
        session : ClientSession
            HTTP session to use.
        request : Resource
            Resource to request.

        Returns
        -------
        Response
            The server's response.
        """
        _request = session.build_request(
            method=request.method,
            url=request.url,
            data=request.data,
            json=request.json,
            params=request.params,
            headers=request.headers,
            timeout=request.timeout,
        )

        try:
            LOGGER.debug("Sending request")
            # perform the request
            response = await session.send(_request, follow_redirects=session.follow_redirects)

            # read the response content
            content = await response.aread()

            try:
                # try to decode the content using the response encoding
                text = content.decode(response.encoding or "utf-8", errors="replace")
            except (LookupError, TypeError):
                # LookupError is raised if the encoding was not found which could
                # indicate a misspelling or similar mistake
                #
                # TypeError can be raised if encoding is None
                encoding = chardet.detect(content)["encoding"]
                text = content.decode(encoding or "utf-8", errors="replace")

            return Response(
                status=response.status_code,
                reason=response.reason_phrase,
                ok=response.is_success,
                text=text,
                content=content,
                url=str(response.url),
            )
        except Exception as error:
            if self._log_errors:
                LOGGER.error(error)

            if self._storage is not None:
                cached_request = httpcore.Request(
                    method=_request.method,
                    url=str(_request.url),
                    headers=_request.headers,  # type: ignore
                    content=_request.content,
                )
                cached_response = httpcore.Response(
                    status=0,
                    content=b"",
                    extensions={"reason_phrase": repr(error).encode("utf-8")},
                )
                cached_response.read()

                await self._storage.store(
                    key=generate_key(cached_request, _request.content),
                    request=cached_request,
                    response=cached_response,
                )

            return Response(status=0, reason=repr(error), ok=False, text="", url="", content=b"")
